import unittest
import unittest.mock

class BreakpointManagerCreated(unittest.TestCase):
	def setUp(self):
		test_breakpoints = [ "init_runlengthdecode_parser", "init_LZW_parser" ]
		self.bpm = BreakpointManager(test_breakpoints)
		self.arch = gdb.selected_frame().architecture()

	def test_locate_retqs(self):
		retqs = self.bpm.locate_retqs("init_runlengthdecode_parser")
		addr = retqs[0]
		instr = self.arch.disassemble(addr, addr+8, 1)[0]
		self.assertTrue(instr['asm'].startswith("ret"))

	def test_locate_retq(self):
		retq = self.bpm.locate_retq("init_runlengthdecode_parser")
		instr = self.arch.disassemble(retq, retq+8, 1)[0]
		self.assertTrue(instr['asm'].startswith("ret"))


class BreakpointManagerSettingBreakpoints(unittest.TestCase):
	def setUp(self):
		#TODO: make these tests independent of the pdf parser
		test_breakpoints = [ "init_runlengthdecode_parser", "init_LZW_parser" ]
		# "constants" like rld_retq below could be moved to setUpClass (possibly better performance)
		self.bpm = BreakpointManager(test_breakpoints)
		self.arch = gdb.selected_frame().architecture()
		self.rld_retq = self.bpm.locate_retq(test_breakpoints[0])
		self.lzw_retq = self.bpm.locate_retq(test_breakpoints[1])

		self.hrbp_patcher = unittest.mock.patch('__main__.HRuleBreakpoint')
		self.hrbp_mock_object = self.hrbp_patcher.start()
		self.plprbp_patcher = unittest.mock.patch('__main__.PerformLowLevelParseRetBreakpoint')
		self.plprbp_mock_object = self.plprbp_patcher.start()
		self.hpprbp_patcher = unittest.mock.patch('__main__.HPackratParseRetBreakpoint')
		self.hpprbp_mock_object = self.hpprbp_patcher.start()

	def tearDown(self):
		self.hpprbp_patcher.stop()
		self.plprbp_patcher.stop()
		self.hrbp_patcher.stop()

# TODO: mock breakpoints, assert on arguments to constructor
	def test_set_h_rule_breakpoints(self):
		self.bpm.set_h_rule_breakpoints()
		self.assertTrue(self.hrbp_mock_object.call_count == 2)
		#bps_valid = [ bp.is_valid() for bp in self.bpm.h_rule_breakpoints]
		# Assert that we have as many breakpoints as h_rule_function and is_valid() returns true for all of them
		#self.assertEqual(bps_valid, len(self.bpm.H_RULE_FUNCTIONS) * [True])
		# Cleanup
		#for bp in self.bpm.h_rule_breakpoints:
		#	bp.delete()

	def test_set_hammer_retq_breakpoints(self):
		self.bpm.set_hammer_retq_breakpoints()
		self.assertTrue(self.plprbp_mock_object.called)
		self.assertTrue(self.hpprbp_mock_object.called)
		#bps_valid = [ bp.is_valid() for bp in self.bpm.hammer_retq_breakpoints]
		#self.assertEqual(bps_valid, 2 * [True])
		#Cleanup
		#for bp in self.bpm.hammer_retq_breakpoints:
		#	bp.delete()

	def test_del_hammer_retq_breakpoints(self):
		self.bpm.set_hammer_retq_breakpoints()
		bps = self.bpm.hammer_retq_breakpoints
		self.bpm.del_hammer_retq_breakpoints()
		for bp in bps:
			self.assertTrue(bp.delete.called)
		#self.assertTrue(self.plprbp_mock_object.delete.called)
		#self.assertTrue(self.hpprbp_mock_object.delete.called)
		#bps_valid = [ bp.is_valid() for bp in bps]
		#self.assertEqual(bps_valid, 2 * [False])

	# TODO
	#def tearDown(self):
		#self.bpm.delete

#TODO: tests for PDFMainBreakpoint
